/*
 *  arch/i386/kernel/rt_time.c
 *
 *  1995-1998, Michael Barabanov
 *
 */


#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/rtl_time.h>
#include <asm/rtl_sync.h>
#include <linux/sched.h>
#include <linux/timex.h>

#include <linux/cons.h>
#include <linux/mc146818rtc.h>




#define READ_CNT0(var) \
{ var = inb(0x40); var |= (inb(0x40) << 8); }

#define READ_CNT2(var) \
{ var = inb(0x42); var |= (inb(0x42) << 8); }


#define LATCH_CNT0() \
outb(0xd2, 0x43);

#define LATCH_CNT0_AND_2() \
outb(0xda, 0x43);

#define LATCH_CNT2() \
outb(0xd8, 0x43);

#define WRITE_COUNTER_ZERO16(x)  { \
	outb_p(x&0xff,0x40); outb_p((x>>8)&0xff,0x40);\
      	timer_counter =x; \
} 

#define WRITE_COUNTER_ZERO8(x)  { \
	outb_p(x&0xff,0x40); \
	timer_counter =x; \
}

#define WRITE_COUNTER_ZERO_ONESHOT(x) { \
	if (x <= 0xff) { \
		if (max_counter0 != 0xff) { \
			outb_p(0x10, 0x43);       /* 8254, channel 0, mode 0, lsb */ \
			max_counter0 = 0xff; \
		} \
		WRITE_COUNTER_ZERO8(x); \
	} else { \
		if (max_counter0 != 0xffff) { \
			outb_p(0x30, 0x43);        /* 8254, channel 0, mode 0, lsb+msb */ \
			max_counter0 = 0xffff; \
		} \
		WRITE_COUNTER_ZERO16(x); \
	} \
}


static unsigned int timer_counter;

static RTIME rt_callback_time;  /* time to invoke the supplied procedure      */
static RTIME rt_intr_time;
enum rtl_timer_modes {PERIODIC, ONESHOT} rtl_timer_mode;
static int max_counter0; // 0xff or 0xffff

static void (*rt_timer_handler)(void);

#define LATCH2 0xfffe

static int can_change_latch2;

RTIME rt_linux_clock;		/* time to generate the next linux interrupt  */

static RTIME rt_time;		/* last time counter 2 was equal to last_c2 */
static int last_c2;

static int first_time_mode2;  /* on changing modes from 0 to 2 the 8254 will generate an interrupt immediately; we skip this interrupt */

static unsigned long scaler;


static void rt_set_latch2(int latch)
{
	long flags;
	if (can_change_latch2) {
		rtl_hard_save_flags(flags);
		rtl_hard_cli();
		outb_p(latch & 0xff, 0x42);
		outb((latch >> 8) & 0xff, 0x42);
		rtl_hard_restore_flags(flags);
	}
}



RTIME rt_get_time_8254(void)
{
	register int c2;
	int flags;

	rtl_hard_save_flags(flags);
	rtl_hard_cli();
	LATCH_CNT2();
	READ_CNT2(c2);
	rt_time += ((c2 < last_c2) ? (last_c2 - c2) / 2 : (last_c2 - c2 + LATCH2) / 2);
	last_c2 = c2;
	rtl_hard_restore_flags(flags);
	return rt_time;
}


RTIME rt_get_time_pentium(void)
{

	RTIME t;

	/* time = counter * scaler / 2^32; */
	/* no div instructions, so it's very fast */
	__asm__("rdtsc\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t) : "b" (scaler) : "cx");

	return t;
}


static void rt_calibrate_pentium(void)
{
	RTIME calibr_time;
	RTIME calibr_time_pentium;

	long flags;

	rtl_hard_save_flags(flags);
	printk("Calibrating pentium counter...");
	rtl_hard_cli();

	scaler = 0x80000000;
	calibr_time_pentium = rt_get_time_pentium();
	calibr_time = rt_get_time_8254();

	while (rt_get_time_8254() < calibr_time + RT_TICKS_PER_SEC);

	calibr_time_pentium = rt_get_time_pentium() - calibr_time_pentium;
	calibr_time = rt_get_time_8254() - calibr_time;

	while (calibr_time_pentium > 0x7ffffff) {
		calibr_time_pentium >>= 1;
		calibr_time >>= 1;
	}

	/* scaler = tim * 2^32 / tim_pentium */
	__asm__("xor %%eax, %%eax\n\t"
		"div %%ecx\n\t"
		:"=a" (scaler) : "d" ((unsigned long)calibr_time), "c" ((unsigned long)calibr_time_pentium));


	scaler /= 2;

	printk("ok. Scaler = %d\n", (int) scaler);
	rtl_hard_restore_flags(flags);
}


RTIME (*do_rt_get_time)(void);

static void rt_reset_timer(void)
{
	int temp;
	int flags;

	if (rtl_timer_mode != ONESHOT) {
		return;
	}

	rtl_hard_save_flags(flags);
	rtl_hard_cli();

	rt_intr_time = rt_get_time();
	if (rt_linux_clock + 300 < rt_callback_time) {		/* there's no RT-callback within 300 time units from the linux clock */
		temp = rt_linux_clock - rt_intr_time;
	}
	else {
		temp = rt_callback_time - rt_intr_time;
	}

	if (temp < 1) {
		temp = 1; 
	}

//	WRITE_COUNTER_ZERO16(temp);
	WRITE_COUNTER_ZERO_ONESHOT(temp);
	rt_intr_time += temp;

	rtl_hard_restore_flags(flags);
}


void rt_oneshot_timer_irq(void)
{
/* 	rt_intr_time = rt_get_time(); */
	if  (rt_intr_time >= rt_linux_clock) {
		rt_linux_clock += LATCH;
		rtl_global_pend_irq(0);
		if (rt_intr_time >= rt_linux_clock) {
			conpr("lost jiffies ");
			conpr("\n");
			rt_linux_clock = rt_intr_time;
		}
	}

	if (rt_intr_time >= rt_callback_time) {
		(*rt_timer_handler)();
	} else {
		rt_reset_timer();
	}
}

void rt_periodic_timer_irq(void)
{
	if  (rt_intr_time >= rt_linux_clock) {
		rt_linux_clock += LATCH;
		rtl_global_pend_irq(0);
	}

	if (rt_timer_handler && !first_time_mode2) {
		(*rt_timer_handler)();   /* should periodically call rt_get_time() */
	} else {
		rt_get_time();		 /* have to keep the clock running */
	}

	first_time_mode2 = 0;

	rt_intr_time += timer_counter;
}

void rtl_set_timer(RTIME when) 
{
	rt_callback_time = when;
	rt_reset_timer();
}


void rtl_no_timer(void) 
{
	rt_callback_time = RT_TIME_END;
	rt_reset_timer();
}


int rtl_request_timer(void (*fn)(void)) 
{
	if (!fn) {
		return -EINVAL;
	}
	if (rt_timer_handler) {
		return -EBUSY;
	}

	rt_timer_handler = fn;
	return 0;
}


void rtl_free_timer(void)
{
	rt_timer_handler = 0;
	rtl_no_timer();
}




void rtl_time_init(void)
{
	int flags;

	rtl_hard_save_flags(flags);
	rtl_hard_cli();

	rt_timer_handler = 0;
	rt_callback_time = RT_TIME_END;
	rt_time = 0;

	do_rt_get_time = rt_get_time_8254;
	can_change_latch2 = 0;

	/* program channel 2 of the 8254 chip for periodic counting */
	outb_p(0xb6, 0x43);	/* binary, mode 3, LSB/MSB, ch 2 */
	outb_p(LATCH2 & 0xff, 0x42);
	outb((LATCH2 >> 8) & 0xff, 0x42);

	LATCH_CNT2();
	READ_CNT2(last_c2);

	outb((inb_p(0x61) & 0xfd) | 1, 0x61); /* shut up the speaker and enable counting */


	if (boot_cpu_data.x86_capability & X86_FEATURE_TSC) {
		  rt_calibrate_pentium(); 
		  can_change_latch2 = 1;
		  do_rt_get_time = rt_get_time_pentium;
	}

	rt_linux_clock = rt_get_time() + LATCH/2;
	rtl_timer_mode = -1;
	I8253_make_oneshot(LATCH);

	rtl_hard_restore_flags(flags);
}

/* returns the start time of the periodic interrupt series */
/* this value can be used as a start_time of a periodic task */

void I8253_make_periodic(unsigned int period)
{
	int flags;
	rtl_hard_save_flags(flags);
	rtl_hard_cli();

	if (period > RT_MAX_PERIOD) {
		period = RT_MAX_PERIOD;
		conpr("RT_MAX_PERIOD exceeded\n");
	}

	rtl_free_global_irq(0);
	rtl_request_global_irq(0,rt_periodic_timer_irq);

	if (rtl_timer_mode == ONESHOT) {
		first_time_mode2 = 1; /* don't pass the first interrupt to the scheduler */
		rtl_timer_mode = PERIODIC;
	}
	outb_p(0x34,0x43);		/* binary, mode 2, LSB/MSB, ch 0 */
	WRITE_COUNTER_ZERO16(period);
	rt_intr_time = rt_get_time();

	rtl_hard_restore_flags(flags);
}

void I8253_make_oneshot(unsigned int d)
{
	int flags;
	if (rtl_timer_mode != ONESHOT)  {
		rtl_hard_save_flags(flags);
		rtl_hard_cli();
		rtl_free_global_irq(0);
		rtl_request_global_irq(0, rt_oneshot_timer_irq);
		rtl_timer_mode = ONESHOT;
		outb_p(0x30, 0x43);       /* 8254, channel 0, mode 0, lsb+msb */
		max_counter0 = 0xffff;
		rt_reset_timer();
		rtl_hard_restore_flags(flags);
	}
}


void I8253_disable_speaker(void)
{
	int flags;
	rtl_hard_save_flags(flags);
	rtl_hard_cli();
	outb(inb_p(0x61) & 0xfd, 0x61);
	rtl_hard_restore_flags(flags);
}

void I8253_enable_speaker(int count)
{
	/* RTL enable speaker. Vanilla Linux thinks timer 2 is off
	   but we do not. */
	int flags;

	rtl_hard_save_flags(flags);
	rtl_hard_cli();

	outb_p(inb_p(0x61)|2, 0x61);
	rt_set_latch2(count);

	rtl_hard_restore_flags(flags);
}

